package co.yixiang.yshop.module.product.service.productcouponrelation;


import co.yixiang.yshop.framework.mybatis.core.query.LambdaQueryWrapperX;
import co.yixiang.yshop.module.product.controller.app.couponrelation.vo.CartCouponVO;
import co.yixiang.yshop.module.product.controller.app.couponrelation.vo.UserCouponVO;
import co.yixiang.yshop.module.product.convert.storeproductcouponrelation.StoreProductCouponRelationConvert;
import co.yixiang.yshop.module.product.dal.dataobject.productcoupon.ProductCouponDO;
import co.yixiang.yshop.module.product.dal.dataobject.productcouponrelation.ProductCouponRelationDO;
import co.yixiang.yshop.module.product.dal.mysql.productcoupon.ProductCouponMapper;
import co.yixiang.yshop.module.product.dal.mysql.productcouponrelation.ProductCouponRelationMapper;
import co.yixiang.yshop.module.product.enums.coupon.CouponExpirationEnum;
import co.yixiang.yshop.module.product.enums.coupon.CouponReceiveTypeEnum;
import co.yixiang.yshop.module.product.enums.coupon.CouponScopeEnum;
import co.yixiang.yshop.module.product.enums.coupon.CouponTypeEnum;
import co.yixiang.yshop.module.product.enums.couponrelation.CouponIsUsedEnum;
import co.yixiang.yshop.module.product.enums.couponrelation.CouponSearchTypeEnum;
import co.yixiang.yshop.module.product.service.productcouponrelation.dto.CartCouponDto;
import com.baomidou.mybatisplus.core.conditions.update.LambdaUpdateWrapper;
import com.baomidou.mybatisplus.core.toolkit.CollectionUtils;
import com.baomidou.mybatisplus.extension.service.impl.ServiceImpl;
import org.springframework.stereotype.Service;
import org.springframework.validation.annotation.Validated;

import javax.annotation.Resource;
import java.math.BigDecimal;
import java.time.LocalDateTime;
import java.time.temporal.ChronoUnit;
import java.util.*;
import java.util.stream.Collectors;

import static co.yixiang.yshop.framework.common.exception.util.ServiceExceptionUtil.exception;
import static co.yixiang.yshop.framework.security.core.util.SecurityFrameworkUtils.getLoginUserId;
import static co.yixiang.yshop.module.product.enums.ErrorCodeConstants.*;

/**
 * App优惠券记录 Service 接口
 *
 * @author moxiangrong
 */
@Service
@Validated
public class AppCouponRelationServiceImpl extends ServiceImpl<ProductCouponRelationMapper, ProductCouponRelationDO> implements AppCouponRelationService {

    @Resource
    private ProductCouponMapper couponMapper;

    @Override
    public void receiveCoupon(Long id) {
        Long uid = getLoginUserId();
        // 查询优惠券
        ProductCouponDO couponDO = couponMapper.selectById(id);
        if (couponDO == null) {
            throw exception(PRODUCT_COUPON_NOT_EXISTS);
        }
        long count = count(new LambdaQueryWrapperX<ProductCouponRelationDO>()
                .eq(ProductCouponRelationDO::getCouponId, id).eq(ProductCouponRelationDO::getUid, uid));

        if (couponDO.getNumber() < 1) throw exception(NOT_PRODUCT_COUPON_ERROR);
        if (CouponReceiveTypeEnum.UNLIMITED.getValue().equals(couponDO.getReceiveType())
                || count < couponDO.getLimitNumber()) {
            ProductCouponRelationDO relationDO = ProductCouponRelationDO.builder().couponId(id).uid(uid).build();
            relationDO.setReceiveTime(LocalDateTime.now());
            save(relationDO);
            couponMapper.updateById(ProductCouponDO.builder().id(couponDO.getId()).number(couponDO.getNumber() - 1).build());
        } else {
            throw exception(PRODUCT_COUPON_RECEIVE_ERROR);
        }
    }

    @Override
    public void verificationCoupon(Long id) {
        ProductCouponRelationDO productCouponDO = getById(id);
        if (productCouponDO == null) {
            throw exception(PRODUCT_COUPON_NOT_EXISTS);
        }
        // 使用优惠券
        productCouponDO.setIsUsed(CouponIsUsedEnum.USED.getValue());
        productCouponDO.setUsageTime(LocalDateTime.now());
        updateById(productCouponDO);
    }

    @Override
    public void returnCoupon(Long id) {
        // 重置优惠券状态
        update(new LambdaUpdateWrapper<ProductCouponRelationDO>()
                .set(ProductCouponRelationDO::getIsUsed, CouponIsUsedEnum.NOT_USED.getValue())
                .set(ProductCouponRelationDO::getUsageTime, null)
                .eq(ProductCouponRelationDO::getId, id));
    }

    @Override
    public List<UserCouponVO> searchUserCoupon(Integer type, Long id) {
        List<UserCouponVO> userCouponVOS = this.baseMapper.searchUserCoupon(type, getLoginUserId(), id);
        for (UserCouponVO userCouponVO : userCouponVOS) {
            if (CouponExpirationEnum.BY_DAY.getValue().equals(userCouponVO.getExpirationType())) {
                // 计算失效时间
                LocalDateTime expirationTime =
                        userCouponVO.getReceiveTime().plus(userCouponVO.getExpirationDay(), ChronoUnit.DAYS);
                userCouponVO.setTakingEffectTime(userCouponVO.getReceiveTime());
                userCouponVO.setExpirationTime(expirationTime);
            }
        }
        return userCouponVOS;
    }

    @Override
    public List<CartCouponVO> searchCartCoupon(List<CartCouponDto> cartCouponDtoList, Long id) {
        List<CartCouponVO> cartCouponVOList = new ArrayList<>();
        // 获取购物车中商品id集合
        Set<Long> productIds =
                cartCouponDtoList.stream().map(CartCouponDto::getProductId).collect(Collectors.toSet());
        Map<Long, BigDecimal> priceMap =
                cartCouponDtoList.stream().collect(Collectors.toMap(CartCouponDto::getProductId, CartCouponDto::getPrice));
        // 查询所有可使用的优惠券
        List<UserCouponVO> userCouponVOList = searchUserCoupon(CouponSearchTypeEnum.CAN_USE.getValue(), id);
        if (CollectionUtils.isEmpty(userCouponVOList)) {
            return cartCouponVOList;
        }
        cartCouponVOList = StoreProductCouponRelationConvert.INSTANCE.convert(userCouponVOList);
        // 过滤出可以使用的优惠券并计算优惠金额
        cartCouponVOList = cartCouponVOList.stream().filter(vo -> filter(productIds, vo, priceMap)).collect(Collectors.toList());
        // 优惠金额降序
        cartCouponVOList = cartCouponVOList.stream()
                .sorted(Comparator.comparing(CartCouponVO::getDiscountAmount).reversed()).collect(Collectors.toList());
        return cartCouponVOList;
    }


    /**
     * 过滤出可以使用的优惠券
     *
     * @param productIds   商品id集合
     * @param cartCouponVO 优惠券信息
     * @param priceMap     金额
     * @return
     */
    private Boolean filter(Set<Long> productIds, CartCouponVO cartCouponVO, Map<Long, BigDecimal> priceMap) {
        if (CouponScopeEnum.PRODUCT.getValue().equals(cartCouponVO.getCouponScope())) {
            Set<Long> ids = new HashSet<>();
            for (String s : cartCouponVO.getScopeValues().split(",")) ids.add(Long.parseLong(s));
            // 取交集
            ids.retainAll(productIds);
            if (CollectionUtils.isEmpty(ids)) return false;
            return compute(ids, cartCouponVO, priceMap);
        } else {
            return compute(productIds, cartCouponVO, priceMap);
        }
    }

    /**
     * 计算优惠金额
     *
     * @param productIds   商品id集合
     * @param cartCouponVO 优惠券信息
     * @param priceMap     金额
     * @return
     */
    private Boolean compute(Set<Long> productIds, CartCouponVO cartCouponVO, Map<Long, BigDecimal> priceMap) {
        // 累计金额
        BigDecimal amount = BigDecimal.ZERO;
        for (Long id : productIds) amount = amount.add(priceMap.get(id));
        // 是否满足门槛
        if (amount.compareTo(cartCouponVO.getThreshold()) < 0) return Boolean.FALSE;
        // 计算优惠金额
        if (CouponTypeEnum.FULL_REDUCTION.getValue().equals(cartCouponVO.getCouponType())) {
            // 优惠金额不可超出商品总价
            if (amount.compareTo(cartCouponVO.getCouponValue()) < 0) {
                cartCouponVO.setDiscountAmount(amount);
            } else {
                cartCouponVO.setDiscountAmount(cartCouponVO.getCouponValue());
            }
        } else {
            cartCouponVO.setDiscountAmount(amount.subtract(amount.multiply(cartCouponVO.getDiscount()).divide(BigDecimal.TEN)));
        }
        return Boolean.TRUE;
    }


}
